/*
 * Coords.cpp
 *
 *  Created on: 18 mars 2017
 *      Author: davidroussel
 */

#include <limits>	// pour numeric_limits
#include <cmath>	// pour les fcts mathématiques
#include <iostream>

#include "Coords.h"

/*
 * Constante utilisée pour la comparaison des valeurs (à epsilon près)
 */
template<class T, size_t N>
const T Coords<T, N>::epsilon = numeric_limits<T>::round_error();

/*
 * Vérifie si au moins l'une des valeurs de la coordonnée est infinie
 * @return true si l'une au moins des valeurs de la coordonnée
 * est infinie.
 * @see #operator *=(const Coords<T, N> & c) throw (CoordsException)
 * @see #operator /=(const Coords<T, N> & c) throw (CoordsException)
 * @see #operator *(const Coords<T, N> & c) const throw (CoordsException)
 * @see #operator /(const Coords<T, N> & c) const throw (CoordsException)
 * @see #dotProduct(const Coords<T, N> & c) const throw (CoordsException)
 * @see #scale(const T & scale) throw (CoordsException)
 */
template<typename T, size_t N>
bool Coords<T, N>::hasInfinity() const
{
	if (numeric_limits<T>::has_infinity)
	{
		for (size_t i = 0; i < N; i++)
		{
			if (fabs(static_cast<double>(data[i])) == numeric_limits<T>::infinity())
			{
				return true;
			}
		}
	}

	return false;
}

/*
 * Constructeur par défaut
 * @post toutes les valeurs de la coordonée ont été initialisées à
 * static_cast<T>(0) <==> T(0).
 */
template<typename T, size_t N>
Coords<T, N>::Coords() // TODO Compléter la liste d'initialisation
{
	/*
	 * Toutes les valeurs de la coordonnées sont initialisées
	 * à leur valeur par défaut T() dans la liste d'initialisation implicite
	 * Seulement on ne peut pas être sûr que T() corresponde bien à T(0),
	 * Il faut donc initialiser toutes les valeurs de data à T(0)
	 */
	// TODO Compléter ...
}

/*
 * Constructeur de copie
 * @param c les coordonnées à copier
 */
template<typename T, size_t N>
Coords<T, N>::Coords(const Coords<T, N>& c) // TODO Compléter la liste d'initialisation
{
	// TODO Compléter ...
}

/*
 * Constructeur à partir d'une initializer_list
 * voir cours : Patrons de classe - Exemple (3/6), p 188
 * @param list la liste des paramètres,
 * peut avoir une longueur différente de N :
 * 	- Si > N, seules les N premières valeurs seront utilisées pour
 * 	affecter les valeurs de la coordonée
 * 	- Si < N, les valeurs contenues dans la liste d'initialisation sont
 * 	toutes utilisées et les valeurs restantes de la coordonnée sont
 * 	initialisées à T(0).
 * @code
 * 	Coord<int, 4> c1 = {1, 2, 3, 4}; // c1 == (1, 2, 3, 4)
 * 	Coord<int, 4> c2 = {2, 3, 4};	// c2 = (2, 3, 4, 0)
 * @endcode
 */
template<typename T, size_t N>
Coords<T, N>::Coords(std::initializer_list<T> list) // TODO Compléter la liste d'initialisation
{
	// TODO Compléter ...
}

/*
 * Destructeur
 */
template<typename T, size_t N>
Coords<T, N>::~Coords()
{
	// TODO evt remettre les coords à zéro (optionnel)
}

/*
 * Accesseur en lecture à la ième coordonnée
 * @param i l'index de la coordonnée voulue
 * @return la ième coordonnée.
 * @throws CoordException si l'indice est invalide avec comme
 * cause CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS
 */
template<typename T, size_t N>
T Coords<T, N>::get(const size_t i) const throw (CoordsException)
{
	// TODO Remplacer par l'implémentation ...
	return static_cast<T>(0);
}

/*
 * Accesseur en lecture/écriture à la ième coordonnée
 * @param i l'index de la coordonnée voulue
 * @return une référence vers la ième coordonnée.
 * @throws CoordException si l'indice est invalide avec comme
 * cause CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS
 */
template<typename T, size_t N>
T& Coords<T, N>::get(const size_t i) throw (CoordsException)
{
	// TODO Remplacer par l'implémentation ...
	return data[0];
}

/*
 * Opérateur d'accès direct en lecture à la ième coordonnée
 * @param i l'index de la coordonnée voulue
 * @return la ième coordonnée.
 * @throws CoordException si l'indice est invalide avec comme
 * cause CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS
 * @see #get(const size_t i) const throw (CoordsException)
 */
template<typename T, size_t N>
T Coords<T, N>::operator [](const size_t i) const throw (CoordsException)
{
	// TODO Remplacer par l'implémentation sans reécrire la même chose que
	// dans get(i) const...
	return static_cast<T>(0);
}

/*
 * Opérateur d'accès direct en lecture/écriture à la ième coordonnée
 * @param i l'index de la coordonnée voulue
 * @return une référence vers la ième coordonnée.
 * @throws CoordException si l'indice est invalide avec comme
 * cause CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS
 * @see #get(const size_t i) throw (CoordsException);
 */
template<typename T, size_t N>
T & Coords<T, N>::operator [](const size_t i) throw (CoordsException)
{
	// TODO Remplacer par l'implémentation sans reécrire la même chose que
	// dans get(i) ...
	return data[0];
}

/*
 * Accesseur en lecture du nombre de valeurs dans ce type de coordonnée
 * @return le nombre de valeurs dans cette coordonnée
 */
template<typename T, size_t N>
size_t Coords<T, N>::size()
{
	return N;
}

/*
 * Opérateur de comparaison
 * @param c la coordonnée à comparer à soi-même
 * @return true si l'écart entre chaque paire de valeurs de la coordonnée
 * courante et de la coordonnée c est inférieur à epsilon
 */
template<typename T, size_t N>
bool Coords<T, N>::operator ==(const Coords<T, N>& c) const
{
	// TODO Remplacer par l'implémentation ...
	return false;
}

/*
 * Opérateur de différence
 * @param c la coordonnée à comparer à soi-même
 * @return false si la distance entre la Coord courante et c est
 * inférieure à epsilon
 * @see #operator ==(const Coords<T, N> & c) const
 */
template<typename T, size_t N>
bool Coords<T, N>::operator !=(const Coords<T, N> & c) const
{
	// TODO Remplacer par l'implémentation sans reécrire la même chose
	// (à 1 lettre près) que dans operator == ...
	return true;
}

/*
 * Opérateur de self incrémentation
 * @param c la coordonnée utilisée pour incrémenter les valeurs de la
 * coordonnée courante
 * @return une référence vers soi-même
 */
template<typename T, size_t N>
Coords<T, N>& Coords<T, N>::operator +=(const Coords<T, N>& c)
{
	// TODO Compléter ...
	return *this;
}

/*
 * Opérateur de self décrémentation
 * @param c la coordonnée utilisée pour décrémenter les valeurs de la
 * coordonnée courante
 * @return une référence vers soi-même
 */
template<typename T, size_t N>
Coords<T, N>& Coords<T, N>::operator -=(const Coords<T, N>& c)
{
	// TODO Compléter ...
	return *this;
}

/*
 * Opérateur de self multiplication
 * @param c la coordonnée utilisée pour multiplier les valeurs de la
 * coordonnée courante
 * @return une référence vers soi-même
 * @throws CoordException si le résultat d'une multiplication produit un
 * nombre infini avec comme cause
 * CoordsException::ExceptionCause::INFINITY
 * @see hasInfinity() const
 */
template<typename T, size_t N>
Coords<T, N>& Coords<T, N>::operator *=(const Coords<T, N>& c)
    throw (CoordsException)
{
	// TODO Compléter ...
	return *this;
}

/*
 * Opérateur de self division
 * @param c la coordonnée utilisée pour diviser les valeurs de la
 * coordonnée courante
 * @return une référence vers soi-même
 * @throws CoordException si le résultat d'une division produit un
 * nombre infini avec comme cause
 * CoordsException::ExceptionCause::INFINITY
 * @see hasInfinity() const
 */
template<typename T, size_t N>
Coords<T, N>& Coords<T, N>::operator /=(const Coords<T, N>& c)
    throw (CoordsException)
{
	// TODO Compléter ...
	return *this;
}

/*
 * Opérateur d'addition avec une autre coordonnée
 * @param c l'autre coordonnée à ajouter à la coordonnée courante
 * @return une nouvelle coordonnée contenant la somme des deux
 * @see #Coords(const Coords<T, N> & c);
 * @see #operator +=(const Coords<T, N> & c)
 */
template<typename T, size_t N>
Coords<T, N> Coords<T, N>::operator +(const Coords<T, N> & c) const
{
	// TODO Remplacer par l'implémentation
	return Coords<T, N>(*this);
}

/*
 * Opérateur de soustraction avec une autre coordonnée
 * @param c l'autre coordonnée à ajouter à la coordonnée courante
 * @return une nouvelle coordonnée contenant la somme des deux
 * @see #Coords(const Coords<T, N> & c);
 * @see #operator -=(const Coords<T, N> & c)
 */
template<typename T, size_t N>
Coords<T, N> Coords<T, N>::operator -(const Coords<T, N> & c) const
{
	// TODO Remplacer par l'implémentation
	return Coords<T, N>(*this);
}

/*
 * Opérateur de multiplication avec une autre coordonnée
 * @param c l'autre coordonnée à ajouter à la coordonnée courante
 * @return une nouvelle coordonnée contenant la somme des deux
 * @throws CoordException si le résultat d'une multiplication produit un
 * nombre infini avec comme cause
 * CoordsException::ExceptionCause::INFINITY
 * @see #Coords(const Coords<T, N> & c);
 * @see #operator *=(const Coords<T, N> & c)
 */
template<typename T, size_t N>
Coords<T, N> Coords<T, N>::operator *(const Coords<T, N> & c) const
    throw (CoordsException)
{
	// TODO Remplacer par l'implémentation
	return Coords<T, N>(*this);
}

/*
 * Opérateur de division avec une autre coordonnée
 * @param c l'autre coordonnée à ajouter à la coordonnée courante
 * @return une nouvelle coordonnée contenant la somme des deux
 * @throws CoordException si le résultat d'une division produit un
 * nombre infini avec comme cause
 * CoordsException::ExceptionCause::INFINITY
 * @see #Coords(const Coords<T, N> & c);
 * @see #operator /=(const Coords<T, N> & c)
 */
template<typename T, size_t N>
Coords<T, N> Coords<T, N>::operator /(const Coords<T, N> & c) const
    throw (CoordsException)
{
	// TODO Remplacer par l'implémentation
	return Coords<T, N>(*this);
}

/*
 * Opérateur ami de sortie standard : Affiche (x1, x2, x3, ... xn) sur le
 * flux de sortie out
 * @param out le flux de sortie sur lequel écrire
 * @param c la coordonnée à afficher sur le flux de sortie
 * @return une référence vers le flux de sortie
 */
template<typename T, size_t N>
ostream & operator <<(ostream & out, const Coords<T, N> & c)
{
	out << "(";

	// TODO Compléter ...

	out << ")";
	return out;
}

/*
 * Produit scalaire de la coordonnée courante avec la coordonnée c
 * @param c la coordonnée avec laquelle calculer le produit scalaire
 * \f[
 * \overrightarrow{X} \bullet \overrightarrow{Y} = \sum_{i = 0}^{n-1} x_{i} \times y_{i}
 * \f]
 * @return la valeur du produit scalaire
 * @throws CoordException si le résultat d'une division produit un
 * nombre infini avec comme cause
 * CoordsException::ExceptionCause::INFINITY
 * @see hasInfinity() const
 */
template<typename T, size_t N>
T Coords<T, N>::dotProduct(const Coords<T, N> & c) const throw (CoordsException)
{
	// TODO Remplacer par l'implémentation
	return static_cast<T>(-1);
}

/*
 * Le carré de la norme euclidienne de cette coordonnée
 * \f[
 * \| \overrightarrow{X} \|^{2} = \overrightarrow{X} \bullet \overrightarrow{X}
 * \f]
 * @return la norme euclidienne au carré de cette coordonnée
 * @see #dotProduct(const Coords<T, N> & c) const
 */
template<typename T, size_t N>
T Coords<T, N>::norm2() const
{
	// TODO Rempplacer par l'implémentation
	return static_cast<T>(-1);
}


/*
 * Translation de la coordonnée courante
 * @param c la coordonnée à ajouter à la coordonnée courante
 * @return une référence vers la coordonnée courante translatée.
 * @see operator +=(const Coords<T, N> & c)
 */
template<typename T, size_t N>
Coords<T, N> & Coords<T, N>::translate(const Coords<T, N> & c)
{
	// TODO Compléter ...
	return *this;
}

/*
 * Homothétie de la coordonnée courante
 * @param scale le facteur d'échelle à applique à chaque valeur de la
 * coordonnée
 * @return une référence vers la coordonnée modifiée
 * @throws CoordException si le résultat d'une multiplication produit un
 * nombre infini avec comme cause
 * CoordsException::ExceptionCause::INFINITY
 * @see hasInfinity() const
 */
template<typename T, size_t N>
Coords<T, N> & Coords<T, N>::scale(const T & scale) throw (CoordsException)
{
	// TODO Compléter ...
	return *this;
}

/*
 * Fonction amie pour accéder au contenu d'une Coords<T, N> sans avoir
 * à utiliser les accesseurs dans les tests
 * @param c la coordonnée dont on veut une valeur
 * @param index l'index de la valeur que l'on souhaite
 * @return la valeur contenue dans la coordonnée à l'index index.
 * @throws CoordsExcepetion si l'index est invalide avec comme
 * cause CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS
 */
template<typename T, size_t N>
T dataAccess(const Coords<T, N> & c, const size_t index) throw (CoordsException)
{
	if (index < N)
	{
		return c.data[index];
	}
	else
	{
		throw CoordsException(
		    CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
	}
}

// ----------------------------------------------------------------------------
// Instanciation des templates pour des types spécifiques
// ----------------------------------------------------------------------------

template class Coords<double, 2> ;
template ostream & operator <<(ostream &, const Coords<double, 2> &);
template double dataAccess(const Coords<double, 2> &, const size_t);
template class Coords<double, 3> ;
template ostream & operator <<(ostream &, const Coords<double, 3> &);
template double dataAccess(const Coords<double, 3> &, const size_t);
template class Coords<double, 4> ;
template ostream & operator <<(ostream &, const Coords<double, 4> &);
template double dataAccess(const Coords<double, 4> &, const size_t);

template class Coords<float, 2> ;
template ostream & operator <<(ostream &, const Coords<float, 2> &);
template float dataAccess(const Coords<float, 2> &, const size_t);
template class Coords<float, 3> ;
template ostream & operator <<(ostream &, const Coords<float, 3> &);
template float dataAccess(const Coords<float, 3> &, const size_t);
template class Coords<float, 4> ;
template ostream & operator <<(ostream &, const Coords<float, 4> &);
template float dataAccess(const Coords<float, 4> &, const size_t);

template class Coords<int, 2> ;
template ostream & operator <<(ostream &, const Coords<int, 2> &);
template int dataAccess(const Coords<int, 2> &, const size_t);
template class Coords<int, 3> ;
template ostream & operator <<(ostream &, const Coords<int, 3> &);
template class Coords<int, 4> ;
template int dataAccess(const Coords<int, 3> &, const size_t);
template ostream & operator <<(ostream &, const Coords<int, 4> &);
template int dataAccess(const Coords<int, 4> &, const size_t);
