/*
 * CoordsException.cpp
 *
 *  Created on: Mar 18, 2017
 *      Author: David Roussel
 */

#include <iostream>	// for endl;
#include <sstream>	// for ostringstream

#include "CoordsException.h"

/*
 * Constructeur de CoordsException avec cause de l'exception
 * @param c la cause de l'exception parmis les causes possibles
 * @see ExceptionCause
 */
CoordsException::CoordsException(CoordsException::ExceptionCause e) :
	exception(),
	cause(e)
{
}

/*
 * Constructeur de CoordsException avec cause de l'exception
 * @param c la cause de l'exception parmis les causes possibles
 * @param descr une chaine de caractère décrivant l'exception
 * @see ExceptionCause
 */
CoordsException::CoordsException(CoordsException::ExceptionCause e, const string & descr) :
	exception(),
	cause(e),
	message(descr)
{
}

/*
 * Destructeur de CoordsException
 */
CoordsException::~CoordsException() throw ()
{
	message.clear();
}

/*
 * Message d'explication de la cause de l'exception
 * @return une chaîne de caractères de type C décrivant l'exception
 * @see #CoordsException(CoordsException::ExceptionCause, const char *)
 */
const char * CoordsException::what() const throw ()
{
	const char * initialWhat = exception::what();

	ostringstream output;

	output << initialWhat << " : ";

	output << "CoordsException : ";

	if (message.length() > 0)
	{
		output << message << " : ";
	}

	switch (cause)
	{
		case ExceptionCause::INDICES_OUT_OF_BOUNDS:
			output << "INDICES OUT OF BOUNDS" << endl;
//			output << "US: indices are out of bounds" << endl;
//			output << "FR: indices hors bornes " << endl;
			break;
		default:
			output << "Unknown exception" << endl;
			break;
	}

	return output.str().c_str();
}

/*
 * Accesseur en lecture de la cause de l'exception
 * @return la cause de l'exeption
 */
CoordsException::ExceptionCause CoordsException::getCause() const
{
	return cause;
}
