/*
 * CoordsException.h
 *
 *  Created on: Mar 18, 2017
 *      Author: David Roussel
 */

#ifndef COORDSEXCEPTION_H_
#define COORDSEXCEPTION_H_

#include <string>		// pour les strings
#include <exception>	// pour la classe de base exception
#include <ostream>		// pour les ostreams

using namespace std;

/**
 * Classe d'exceptions pour les Tableau<Type>
 */
class CoordsException: public exception
{
	public:
		/**
		 * Cas d'exception
		 */
		using ExceptionCause = enum
		{
			/**
			 * Indice en dehors des bornes du tableau courant
			 */
			INDICES_OUT_OF_BOUNDS,
			/**
			 * Valeurs infinies
			 */
			INFINITY_VALUE,
			/**
			 * When illegal memory access occurs
			 */
			ILLEGAL_MEMORY_ACCESS,
			/**
			 * When Abort signal occurs
			 */
			ABORT
		};

		/**
		 * Constructeur de CoordsException avec cause de l'exception
		 * @param c la cause de l'exception parmis les causes possibles
		 * @see ExceptionCause
		 */
		CoordsException(CoordsException::ExceptionCause c);

		/**
		 * Constructeur de CoordsException avec cause de l'exception
		 * @param c la cause de l'exception parmis les causes possibles
		 * @param descr une chaine de caractère décrivant l'exception
		 * @see ExceptionCause
		 */
		CoordsException(CoordsException::ExceptionCause e, const string & descr);

		/**
		 * Destructeur de CoordsException
		 */
		virtual ~CoordsException() throw ();

		/**
		 * Message d'explication de la cause de l'exception
		 * @return une chaîne de caractères de type C décrivant l'exception
		 * @see #CoordsException(CoordsException::ExceptionCause, const char *)
		 */
		virtual const char* what() const throw ();

		/**
		 * Accesseur en lecture de la cause de l'exception
		 * @return la cause de l'exeption
		 */
		ExceptionCause getCause() const;

		/*
		 * Note : l'opérateur de sortie standard n'a pas besoin d'être réimp-
		 * lémenté car on utilise habituellement la méthode what() pour
		 * afficher la cause de l'exception.
		 * Exemple:
		 * try
		 * {
		 * 	... faire qq chose qui throw une std::exception
		 * }
		 * catch (exception & e)
		 * {
		 * 	cerr << e.what() << endl;
		 * }
		 *
		 */

	protected:
		/**
		 * La cause de l'exception
		 */
		CoordsException::ExceptionCause cause;

		/**
		 * Le message de description de l'exception
		 */
		string message;

};

#endif /* COORDSEXCEPTION_H_ */
